'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

const last = (v = []) => v[v.length - 1];

const isWhitespace = (node) => {
  return node.nodeType === node.TEXT_NODE && node.nodeValue.trim() === "";
};

const walk = (node, callback, deep = true) => {
  if (!node) return;
  if (!isWhitespace(node)) {
    let v = callback(node);
    if (v === false) return;
    if (v?.nodeName) return walk(v, callback, deep);
  }
  if (deep) walk(node.firstChild, callback, deep);
  walk(node.nextSibling, callback, deep);
};

const transformBrackets = (str) => {
  let parts = str.split(/(\[[^\]]+\])/).filter((v) => v);
  return parts.reduce((a, part) => {
    let v = part.charAt(0) === "[" ? "." + part.replace(/\./g, ":") : part;
    return a + v;
  }, "");
};

const getTarget = (path, target) => {
  let parts = transformBrackets(path)
    .split(".")
    .map((k) => {
      if (k.charAt(0) === "[") {
        let p = k.slice(1, -1).replace(/:/g, ".");
        return getValueAtPath(p, target);
      } else {
        return k;
      }
    });

  let t =
    parts.slice(0, -1).reduce((o, k) => {
      return o && o[k];
    }, target) || target;
  return [t, last(parts)];
};

const callFunctionAtPath = (path, target, args) => {
  let [a, b] = getTarget(path, target);
  return a[b]?.apply?.(a, args);
};

const getValueAtPath = (path, target) => {
  let [a, b] = getTarget(path, target);
  let v = a[b];
  if (typeof v === "function") return v.bind(a);
  return v;
};

const setValueAtPath = (path, value, target) => {
  let [a, b] = getTarget(path, target);
  return (a[b] = value);
};

const fragmentFromTemplate = (v) => {
  if (typeof v === "string") {
    let tpl = document.createElement("template");
    tpl.innerHTML = v.trim();
    return tpl.content;
  }
  if (v.nodeName === "TEMPLATE") return v.cloneNode(true).content;
  if (v.nodeName === "defs") return v.firstElementChild.cloneNode(true);
};

const debounce = (fn) => {
  let t;
  return function (...args) {
    if (t) return;
    t = requestAnimationFrame(() => {
      fn(...args);
      t = null;
    });
  };
};

const isPrimitive = (v) => v === null || typeof v !== "object";

const typeOf = (v) =>
  Object.prototype.toString.call(v).match(/\s(.+[^\]])/)[1];

const pascalToKebab = (string) =>
  string.replace(/[\w]([A-Z])/g, function (m) {
    return m[0] + "-" + m[1].toLowerCase();
  });

const kebabToPascal = (string) =>
  string.replace(/[\w]-([\w])/g, function (m) {
    return m[0] + m[2].toUpperCase();
  });

const applyAttribute = (node, name, value) => {
  name = pascalToKebab(name);

  if (typeof value === "boolean") {
    if (name.startsWith("aria-")) {
      value = "" + value;
    } else if (value) {
      value = "";
    }
  }

  if (typeof value === "string" || typeof value === "number") {
    node.setAttribute(name, value);
  } else {
    node.removeAttribute(name);
  }
};

const attributeToProp = (k, v) => {
  let name = kebabToPascal(k);
  if (v === "") v = true;
  if (k.startsWith("aria-")) {
    if (v === "true") v = true;
    if (v === "false") v = false;
  }
  return {
    name,
    value: v,
  };
};

exports.applyAttribute = applyAttribute;
exports.attributeToProp = attributeToProp;
exports.callFunctionAtPath = callFunctionAtPath;
exports.debounce = debounce;
exports.fragmentFromTemplate = fragmentFromTemplate;
exports.getValueAtPath = getValueAtPath;
exports.isPrimitive = isPrimitive;
exports.isWhitespace = isWhitespace;
exports.kebabToPascal = kebabToPascal;
exports.last = last;
exports.pascalToKebab = pascalToKebab;
exports.setValueAtPath = setValueAtPath;
exports.typeOf = typeOf;
exports.walk = walk;
