const TEXT = 1;
const ATTRIBUTE = 2;
const INPUT = 3;
const EVENT = 4;
const REPEAT = 5;

const updateFormControl = (node, value) => {
  if (node.hasAttribute("multiple") && node.nodeName === "SELECT") {
    Array.from(node.querySelectorAll("option")).forEach((option) => {
      option.selected = value.includes(option.value);
    });
    return;
  }

  let checked;

  switch (node.getAttribute("type")) {
    case "checkbox":
      checked = value;
      if (node.checked === checked) break;
      if (checked) {
        node.setAttribute("checked", "");
      } else {
        node.removeAttribute("checked");
      }
      break;
    case "radio":
      checked = value === node.getAttribute("value");
      if (node.checked === checked) break;
      node.checked = checked;
      if (checked) {
        node.setAttribute("checked", "");
      } else {
        node.removeAttribute("checked");
      }
      break;
    default:
      if (node.value === value) break;
      node.setAttribute("value", (node.value = value || ""));
      break;
  }
};

const last = (v = []) => v[v.length - 1];

const isWhitespace = (node) => {
  return node.nodeType === node.TEXT_NODE && node.nodeValue.trim() === "";
};

const walk = (node, callback, deep = true) => {
  if (!node) return;
  if (!isWhitespace(node)) {
    let v = callback(node);
    if (v === false) return;
    if (v?.nodeName) return walk(v, callback, deep);
  }
  if (deep) walk(node.firstChild, callback, deep);
  walk(node.nextSibling, callback, deep);
};

const transformBrackets = (str) => {
  let parts = str.split(/(\[[^\]]+\])/).filter((v) => v);
  return parts.reduce((a, part) => {
    let v = part.charAt(0) === "[" ? "." + part.replace(/\./g, ":") : part;
    return a + v;
  }, "");
};

const getTarget = (path, target) => {
  let parts = transformBrackets(path)
    .split(".")
    .map((k) => {
      if (k.charAt(0) === "[") {
        let p = k.slice(1, -1).replace(/:/g, ".");
        return getValueAtPath(p, target);
      } else {
        return k;
      }
    });

  let t =
    parts.slice(0, -1).reduce((o, k) => {
      return o && o[k];
    }, target) || target;
  return [t, last(parts)];
};

const callFunctionAtPath = (path, target, args) => {
  let [a, b] = getTarget(path, target);
  return a[b]?.apply?.(a, args);
};

const getValueAtPath = (path, target) => {
  let [a, b] = getTarget(path, target);
  let v = a[b];
  if (typeof v === "function") return v.bind(a);
  return v;
};

const setValueAtPath = (path, value, target) => {
  let [a, b] = getTarget(path, target);
  return (a[b] = value);
};

const fragmentFromTemplate = (v) => {
  if (typeof v === "string") {
    let tpl = document.createElement("template");
    tpl.innerHTML = v.trim();
    return tpl.content;
  }
  if (v.nodeName === "TEMPLATE") return v.cloneNode(true).content;
  if (v.nodeName === "defs") return v.firstElementChild.cloneNode(true);
};

const debounce = (fn) => {
  let t;
  return function (...args) {
    if (t) return;
    t = requestAnimationFrame(() => {
      fn(...args);
      t = null;
    });
  };
};

const isPrimitive = (v) => v === null || typeof v !== "object";

const typeOf = (v) =>
  Object.prototype.toString.call(v).match(/\s(.+[^\]])/)[1];

const pascalToKebab = (string) =>
  string.replace(/[\w]([A-Z])/g, function (m) {
    return m[0] + "-" + m[1].toLowerCase();
  });

const kebabToPascal = (string) =>
  string.replace(/[\w]-([\w])/g, function (m) {
    return m[0] + m[2].toUpperCase();
  });

const applyAttribute = (node, name, value) => {
  name = pascalToKebab(name);

  if (typeof value === "boolean") {
    if (name.startsWith("aria-")) {
      value = "" + value;
    } else if (value) {
      value = "";
    }
  }

  if (typeof value === "string" || typeof value === "number") {
    node.setAttribute(name, value);
  } else {
    node.removeAttribute(name);
  }
};

const attributeToProp = (k, v) => {
  let name = kebabToPascal(k);
  if (v === "") v = true;
  if (k.startsWith("aria-")) {
    if (v === "true") v = true;
    if (v === "false") v = false;
  }
  return {
    name,
    value: v,
  };
};

function parseEach(node) {
  let each = node.getAttribute("each");
  let m = each?.match(/(.+)\s+in\s+(.+)/);
  if (!m) {
    if (!each) return m;
    return {
      path: each.trim(),
      key: node.getAttribute("key"),
    };
  }
  let [_, left, right] = m;
  let parts = left.match(/\(([^\)]+)\)/);
  let [a, b] = (parts ? parts[1].split(",") : [left]).map((v) => v.trim());

  return {
    path: right.trim(),
    identifier: b ? b : a,
    index: b ? a : b,
    key: node.getAttribute("key"),
  };
}

const getBlockSize = (template) => {
  let i = 0;
  walk(template.content?.firstChild || template.firstChild, () => i++, false);
  return i;
};

const nextNonWhitespaceSibling = (node) => {
  return isWhitespace(node.nextSibling)
    ? nextNonWhitespaceSibling(node.nextSibling)
    : node.nextSibling;
};

const getBlockFragments = (template, numBlocks) => {
  let blockSize = getBlockSize(template);

  let r = [];
  if (numBlocks) {
    while (numBlocks--) {
      let f = document.createDocumentFragment();
      let n = blockSize;
      while (n--) {
        f.appendChild(nextNonWhitespaceSibling(template));
      }
      r.push(f);
    }
  }
  return r;
};

const getBlocks = (template, numBlocks) => {
  let blockSize = getBlockSize(template);
  let r = [];
  let node = template;
  if (numBlocks) {
    while (numBlocks--) {
      let f = [];
      let n = blockSize;
      while (n--) {
        node = nextNonWhitespaceSibling(node);
        f.push(node);
      }
      r.push(f);
    }
  }
  return r;
};

const compareKeyedLists = (key, a = [], b = []) => {
  let delta = b.map(([k, item]) =>
    !key ? (k in a ? k : -1) : a.findIndex(([_, v]) => v[key] === item[key])
  );
  if (a.length !== b.length || !delta.every((a, b) => a === b)) return delta;
};

function lastChild(v) {
  return (v.nodeType === v.DOCUMENT_FRAGMENT_NODE && v.lastChild) || v;
}

const updateList = (template, delta, entries, createListItem) => {
  let n = template.getAttribute("length") || 0;
  let blocks = getBlockFragments(template, n);
  let t = template;

  delta.forEach((i, newIndex) => {
    let frag =
      i === -1
        ? createListItem(entries[newIndex][1], entries[newIndex][0])
        : blocks[i];
    let x = lastChild(frag);
    t.after(frag);
    t = x;
  });

  template.setAttribute("length", delta.length);
};

const proxy = (root, callback) => {
  let proxyCache = new WeakMap();

  function createProxy(target, handler) {
    let proxy = proxyCache.get(target);
    if (proxy === undefined) {
      proxy = new Proxy(target, handler);
      proxyCache.set(target, proxy);
    }
    return proxy;
  }

  const handler = {
    get(target, property) {
      if (["Object", "Array"].includes(typeOf(target[property]))) {
        return createProxy(target[property], handler);
      } else {
        return Reflect.get(...arguments);
      }
    },
    set(target, property, value) {
      if (value === target[property]) return true;

      callback();
      return Reflect.set(...arguments);
    },
    deleteProperty() {
      callback();
      return Reflect.deleteProperty(...arguments);
    },
  };

  return new Proxy(root, handler);
};

const VALUE = 1;
const KEY = 2;
const FUNCTION = 3;

const hasMustache = (v) => v.match(/({{[^{}]+}})/);

const parseEventHandler = (value) => {
  let m = value.match(/(?:(\w+) => )?([^\(]+)(?:\(([^\)]*)\))/);

  if (!m) return;

  let event = m[1],
    method = m[2],
    args = m[3] ? m[3].split(",").map((v) => v.trim()) : [];

  return {
    event,
    method,
    args,
  };
};

const getParts = (value, context) =>
  value
    .trim()
    .split(/({{[^{}]+}})/)
    .filter((v) => v)
    .map((value) => {
      let match = value.match(/{{([^{}]+)}}/);

      if (!match)
        return {
          type: VALUE,
          value,
          negated: false,
        };

      value = match[1].trim();
      let negated = value.charAt(0) === "!";
      if (negated) value = value.slice(1);

      let fn = parseEventHandler(value);

      return fn
        ? {
            type: FUNCTION,
            negated,
            ...fn,
          }
        : {
            type: KEY,
            value,
            negated,
          };
    });

const getValueFromParts = (target, parts) => {
  return parts.reduce((a, part) => {
    let { type, value, negated } = part;

    let v;

    if (type === VALUE) v = value;
    if (type === KEY) {
      v = getValueAtPath(value, target);
    }
    if (type === FUNCTION) {
      let args = part.args.map((value) => getValueAtPath(value, target));

      v = getValueAtPath(part.method, target)?.(...args);
    }

    if (negated) v = !v;

    return a || a === 0 ? a + v : v;
  }, "");
};

const pascalToKebab$1 = (string) =>
  string.replace(/[\w]([A-Z])/g, function (m) {
    return m[0] + "-" + m[1].toLowerCase();
  });

const kebabToPascal$1 = (string) =>
  string.replace(/[\w]-([\w])/g, function (m) {
    return m[0] + m[2].toUpperCase();
  });

const parseStyles = (value) => {
  let type = typeof value;

  if (type === "string")
    return value.split(";").reduce((o, value) => {
      const [k, v] = value.split(":").map((v) => v.trim());
      if (k) o[k] = v;
      return o;
    }, {});

  if (type === "object") return value;

  return {};
};

const joinStyles = (value) =>
  Object.entries(value)
    .map(([k, v]) => `${k}: ${v};`)
    .join(" ");

const convertStyles = (o) =>
  Object.keys(o).reduce((a, k) => {
    a[pascalToKebab$1(k)] = o[k];
    return a;
  }, {});

const applyAttribute$1 = (node, name, value) => {
  if (name === "style") {
    value = joinStyles(
      convertStyles({
        ...parseStyles(node.getAttribute("style")),
        ...parseStyles(value),
      })
    );
  } else if (name === "class") {
    switch (typeOf(value)) {
      case "Array":
        value = value.join(" ");
        break;
      case "Object":
        value = Object.keys(value)
          .reduce((a, k) => {
            if (value[k]) a.push(k);
            return a;
          }, [])
          .join(" ");
        break;
    }
  } else if (!isPrimitive(value)) {
    return (node[kebabToPascal$1(name)] = value);
  }

  name = pascalToKebab$1(name);

  if (typeof value === "boolean") {
    if (name.startsWith("aria-")) {
      value = "" + value;
    } else if (value) {
      value = "";
    }
  }

  let current = node.getAttribute(name);

  if (value === current) return;

  if (typeof value === "string" || typeof value === "number") {
    node.setAttribute(name, value);
  } else {
    node.removeAttribute(name);
  }
};

const HYDRATE_ATTR = "synergy-hydrate";

const render = (
  target,
  model,
  template,
  updatedCallback,
  beforeMountCallback
) => {
  let observer = () => {
    let subscribers = new Set();
    return {
      publish: (cb) => {
        for (let fn of subscribers) {
          fn();
        }
        cb?.();
      },
      subscribe(fn) {
        subscribers.add(fn);
      },
    }
  };

  const _ = (a, b) => (a === undefined ? b : a);

  const createSubscription = {
    [TEXT]: ({ value, node, model }) => {
      return {
        handler: () => {
          let a = node.textContent;
          let b = getValueFromParts(model, getParts(value));
          if (a !== b) node.textContent = b;
        },
      }
    },
    [ATTRIBUTE]: ({ value, node, model, name }) => {
      return {
        handler: () => {
          let b = getValueFromParts(model, getParts(value));
          applyAttribute$1(node, name, b);
        },
      }
    },
    [INPUT]: ({ node, model, path }) => {
      node.addEventListener("input", () => {
        let value =
          node.getAttribute("type") === "checkbox" ? node.checked : node.value;

        if (value.trim?.().length && !isNaN(value)) value = +value;
        setValueAtPath(path, value, model);
      });
      return {
        handler: () => {
          // let a = node.value;
          let b = getValueAtPath(path, model);
          updateFormControl(node, b);
        },
      }
    },
    [EVENT]: ({ node, model, eventType, event, method, args }) => {
      node.addEventListener(eventType, (e) => {
        let x = args
          ? args.map((v) => (v === event ? e : _(getValueAtPath(v, model), v)))
          : [];

        let fn = callFunctionAtPath(method, model, x);

        if (fn) {
          requestAnimationFrame(fn);
        }
      });
      return {
        handler: () => {},
      }
    },
    [REPEAT]: ({
      node,
      model,
      map,
      path,
      identifier,
      index,
      key,
      blockIndex,
      hydrate,
      pickupNode,
    }) => {
      let oldValue;
      node.$t = blockIndex - 1;

      const initialiseBlock = (rootNode, i, k, exitNode) => {
        let wrapper = new Proxy(model, {
          get(_, property) {
            let x = getValueAtPath(path, model);

            if (property === identifier) {
              for (let n in x) {
                let v = x[n];

                if (key) {
                  if (v[key] === k) return v
                } else {
                  if (n == i) return v
                }
              }
            }

            if (property === index) {
              for (let n in x) {
                let v = x[n];
                if (key) {
                  if (v[key] === k) return n
                } else {
                  if (n == i) return n
                }
              }
            }

            if (x[i]?.hasOwnProperty?.(property)) return x[i][property]

            return Reflect.get(...arguments)
          },
        });

        walk(
          rootNode,
          multi(
            (node) => {
              if (node === exitNode) return false
            },
            bindAll(wrapper, map, hydrate),
            (child) => (child.$t = blockIndex)
          )
        );
      };

      function firstChild(v) {
        return (v.nodeType === v.DOCUMENT_FRAGMENT_NODE && v.firstChild) || v
      }

      const createListItem = (datum, i) => {
        let k = datum[key];
        let frag = fragmentFromTemplate(node);
        initialiseBlock(firstChild(frag), i, k);
        return frag
      };

      if (hydrate) {
        let x = getValueAtPath(path, model);
        let length = typeOf(x) === "Object" ? Object.keys(x).length : x.length;
        let blocks = getBlocks(node, length);
        blocks.forEach((block, i) => {
          let datum = x[i];
          let k = datum[key];
          initialiseBlock(block[0], i, k, last(block).nextSibling);
        });
        pickupNode = last(last(blocks)).nextSibling;
      }

      return {
        handler: () => {
          const newValue = Object.entries(getValueAtPath(path, model) || []);
          const delta = compareKeyedLists(key, oldValue, newValue);

          if (delta) {
            updateList(node, delta, newValue, createListItem);
          }
          oldValue = newValue.slice(0);
        },
        pickupNode,
      }
    },
  };

  const mediator = () => {
    const o = observer();
    return {
      bind(v) {
        let s = createSubscription[v.type](v);
        o.subscribe(s.handler);
        return s
      },
      scheduleUpdate: debounce((cb) => {
        return o.publish(cb)
      }),
      update(cb) {
        return o.publish(cb)
      },
    }
  };

  const { bind, update, scheduleUpdate } = mediator();

  let blockCount = 0;

  const parse = (frag) => {
    let index = 0;
    let map = {};

    walk(frag, (node) => {
      let x = [];
      let pickupNode;
      switch (node.nodeType) {
        case node.TEXT_NODE: {
          let value = node.textContent;
          if (hasMustache(value)) {
            x.push({
              type: TEXT,
              value,
            });
          }
          break
        }
        case node.ELEMENT_NODE: {
          let each = parseEach(node);

          if (each) {
            let ns = node.namespaceURI;
            let m;

            if (ns.endsWith("/svg")) {
              node.removeAttribute("each");
              let tpl = document.createElementNS(ns, "defs");
              tpl.innerHTML = node.outerHTML;
              node.parentNode.replaceChild(tpl, node);
              node = tpl;
              m = parse(node.firstChild);
            } else {
              if (node.nodeName !== "TEMPLATE") {
                node.removeAttribute("each");
                let tpl = document.createElement("template");

                tpl.innerHTML = node.outerHTML;
                node.parentNode.replaceChild(tpl, node);
                node = tpl;
              }
              m = parse(node.content.firstChild);
            }

            pickupNode = node.nextSibling;

            x.push({
              type: REPEAT,
              map: m,
              blockIndex: blockCount++,
              ...each,
              pickupNode,
            });

            break
          }

          let attrs = node.attributes;
          let i = attrs.length;
          while (i--) {
            let { name, value } = attrs[i];

            if (
              name === ":name" &&
              value &&
              (node.nodeName === "INPUT" ||
                node.nodeName === "SELECT" ||
                node.nodeName === "TEXTAREA")
            ) {
              x.push({
                type: INPUT,
                path: value,
              });
            }

            if (name.startsWith(":on")) {
              node.removeAttribute(name);
              let eventType = name.split(":on")[1];
              x.push({
                type: EVENT,
                eventType,
                ...parseEventHandler(value),
              });
            } else if (name.startsWith(":")) {
              let prop = name.slice(1);

              x.push({
                type: ATTRIBUTE,
                name: prop,
                value: value || `{{${prop}}}`,
              });
              node.removeAttribute(name);
            }
          }
        }
      }
      if (x.length) map[index] = x;
      index++;
      return pickupNode
    });

    return map
  };

  const multi =
    (...fns) =>
    (...args) => {
      for (let fn of fns) {
        let v = fn(...args);
        if (v === false) return false
      }
    };

  const bindAll = (model, bMap, hydrate) => {
    let index = 0;
    return (node) => {
      let k = index;
      let p;
      if (k in bMap) {
        bMap[k].forEach((v) => {
          let x = bind({
            ...v,
            node,
            model,
            hydrate,
          });
          p = x.pickupNode;
        });
        node.$i = index;
      }
      index++;
      return p
    }
  };

  let prevState = Object.assign({}, model);
  let p = proxy(model, () =>
    scheduleUpdate(() => {
      updatedCallback?.();
      p.updatedCallback?.(prevState);
      prevState = Object.assign({}, model);
    })
  );
  let frag = fragmentFromTemplate(template);
  let map = parse(frag);
  let hydrate = target.hasAttribute?.(HYDRATE_ATTR);
  if (hydrate) {
    walk(target, bindAll(p, map, 1));
  } else {
    walk(frag, bindAll(p, map));
    beforeMountCallback?.(frag);
    target.appendChild(frag);
    update();
    target.setAttribute?.(HYDRATE_ATTR, 1);
  }

  return p
};

const childNodes = (node) => {
  let frag = document.createDocumentFragment();
  while (node.firstChild) {
    frag.appendChild(node.firstChild);
  }
  return frag;
};

const mergeSlots = (targetNode, sourceNode) => {
  let namedSlots = sourceNode.querySelectorAll("slot[name]");

  namedSlots.forEach((slot) => {
    let name = slot.attributes.name.value;
    let node = targetNode.querySelector(`[slot="${name}"]`);
    if (!node) {
      slot.parentNode.replaceChild(childNodes(slot), slot);
      return;
    }
    node.removeAttribute("slot");
    slot.parentNode.replaceChild(node, slot);
  });

  let defaultSlot = sourceNode.querySelector("slot:not([name])");

  if (defaultSlot) {
    defaultSlot.parentNode.replaceChild(
      childNodes(targetNode.innerHTML.trim() ? targetNode : defaultSlot),
      defaultSlot
    );
  }
};

const define = (name, factory, template, options = {}) =>
  customElements.define(
    name,
    class extends HTMLElement {
      async connectedCallback() {
        if (!this.initialised) {
          this.$viewmodel = factory(this);

          if (this.$viewmodel instanceof Promise)
            this.$viewmodel = await this.$viewmodel;

          let observedProps = Object.keys(this.$viewmodel).filter(
            (v) => v.charAt(0) === "$"
          );

          let observedAttributes = observedProps
            .map((v) => v.slice(1))
            .map(pascalToKebab);

          let sa = this.setAttribute;
          this.setAttribute = (k, v) => {
            if (observedAttributes.includes(k)) {
              let { name, value } = attributeToProp(k, v);
              this.$viewmodel["$" + name] = value;
            }
            sa.apply(this, [k, v]);
          };

          observedAttributes.forEach((name) => {
            let property = attributeToProp(name).name;

            let value;

            if (this.hasAttribute(name)) {
              value = this.getAttribute(name);
            } else {
              value = this[property] || this.$viewmodel["$" + property];
            }

            Object.defineProperty(this, property, {
              get() {
                return this.$viewmodel["$" + property];
              },
              set(v) {
                this.$viewmodel["$" + property] = v;
                if (isPrimitive(v)) {
                  applyAttribute(this, property, v);
                }
              },
            });

            this[property] = value;
          });

          let beforeMountCallback;

          if (options.shadow) {
            this.attachShadow({
              mode: options.shadow,
            });
          } else {
            beforeMountCallback = (frag) => mergeSlots(this, frag);
          }

          this.$viewmodel = render(
            this.shadowRoot || this,
            this.$viewmodel,
            template,
            () => {
              observedProps.forEach((k) => {
                let v = this.$viewmodel[k];
                if (isPrimitive(v)) applyAttribute(this, k.slice(1), v);
              });
            },
            beforeMountCallback
          );
          this.initialised = true;
        }
        this.$viewmodel.connectedCallback?.(this.$viewmodel);
      }
      disconnectedCallback() {
        this.$viewmodel?.disconnectedCallback?.(this.$viewmodel);
      }
    }
  );

export { define };
